# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
network functions utilities for mozregression.
"""

import re
import redo
import requests

from bs4 import BeautifulSoup


def retry_get(url, **karwgs):
    """
    More robust `requests.get` equivalent function.

    This is equivalent to the requests.get function, except that
    it will retry the requests call three times in case of HTTPError or
    ConnectionError.
    """
    return redo.retry(get_http_session().get, attempts=3, sleeptime=1,
                      retry_exceptions=(requests.exceptions.HTTPError,
                                        requests.exceptions.ConnectionError),
                      args=(url,), kwargs=karwgs)


SESSION = None


def set_http_session(session=None, get_defaults=None):
    """
    Define a cache http session.

    :param cache_session: a customized request session or None to use a
                          simple request session.
    :param: get_defaults: if defined, it must be a dict that will provide
        default values for calls to cache_session.get.
    """
    global SESSION
    if get_defaults:
        if session is None:
            session = requests.Session()
        # monkey patch to set default values to a session.get calls
        # I don't see other ways to do this globally for timeout for example
        _get = session.get

        def _default_get(*args, **kwargs):
            for k, v in get_defaults.iteritems():
                kwargs.setdefault(k, v)
            return _get(*args, **kwargs)
        session.get = _default_get
    SESSION = session


def get_http_session():
    """
    Returns the defined http session.
    """
    return SESSION or requests


def url_links(url, regex=None, auth=None):
    """
    Returns a list of links that can be found on a given web page.

    Note that it href of links found are starting with '/' (absolute url from
    the server point of view) then only the last part of the link will be
    returned. E.G:

    '/pub/firefox/nightly/2015/03/2015-03-02-03-02-04-mozilla-central/'
    will return '2015-03-02-03-02-04-mozilla-central/'.
    """
    response = retry_get(url, auth=auth)
    response.raise_for_status()

    soup = BeautifulSoup(response.text)

    if regex:
        if isinstance(regex, basestring):
            regex = re.compile(regex)
        match = regex.match
    else:
        def match(_):
            return True

    # do not return a generator but an array, so we can store it for later use
    result = []
    for link in soup.findAll('a'):
        href = link.get('href')
        # return "relative" part of the url only
        if href.startswith('/'):
            if href.endswith('/'):
                href = href.strip('/').rsplit('/', 1)[-1] + '/'
            else:
                href = href.rsplit('/', 1)[-1]
        if match(href):
            result.append(href)
    return result



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Reading and writing of the configuration file.
"""

import os
import mozinfo

from configobj import ConfigObj, ParseError

from mozregression.log import colorize
from mozregression.errors import MozRegressionError


CONFIG_FILE_HELP_URL = (
    "http://mozilla.github.io/mozregression/documentation/configuration.html"
)
DEFAULT_CONF_FNAME = os.path.expanduser(
    os.path.join("~", ".mozilla", "mozregression", "mozregression.cfg")
)
TC_CREDENTIALS_FNAME = os.path.expanduser(
    os.path.join("~", ".mozilla", "mozregression",
                 "taskcluster-credentials.json")
)
ARCHIVE_BASE_URL = "https://archive.mozilla.org/pub"
# when a bisection range needs to be expanded, the following value is used to
# specify how many builds we try (if 20, we will try 20 before the lower limit,
# and another 20 after the higher limit)
DEFAULT_EXPAND = 20

# default values when not defined in config file.
# Note that this is also the list of options that can be used in config file
DEFAULTS = {
    'adb-profile-dir': None,
    'app': 'firefox',
    'approx-policy': 'auto',
    'archive-base-url': ARCHIVE_BASE_URL,
    'background-dl-policy': 'cancel',
    'bits': None,
    'build-type': '',
    'cmdargs': [],
    'http-timeout': 30.0,
    'inbound-branch': None,
    'mode': 'classic',
    'no-background-dl': '',
    'persist': None,
    'persist-size-limit': 0,
    'process-output': None,
    'profile': None,
    'profile-persistence': 'clone',
    'repo': None,
    'taskcluster-accesstoken': None,
    'taskcluster-clientid': None,
}


def get_defaults(conf_path):
    """
    Get custom defaults from configuration file in argument.
    """
    defaults = dict(DEFAULTS)
    try:
        config = ConfigObj(conf_path)
    except ParseError, exc:
        raise MozRegressionError(
            "Error while reading the config file %s:\n  %s" % (conf_path, exc)
        )
    defaults.update(config)

    return defaults


def _get_persist_dir(default):
    print("You should configure a persist directory, where to put downloaded"
          " build files to reuse them in future bisections.")
    print("I recommend using %s. Leave blank to use that default. If you"
          " really don't want a persist dir type NONE, else you can"
          " just define a path that you would like to use." % default)
    value = raw_input("persist: ")
    if value == "NONE":
        return ''
    elif value:
        persist_dir = os.path.realpath(value)
    else:
        persist_dir = default

    if persist_dir:
        if not os.path.isdir(persist_dir):
            os.makedirs(persist_dir)
    return persist_dir


def _get_persist_size_limit(default):
    print("You should choose a size limit for the persist dir. I recommend you"
          " to use %s GiB, so leave it blank to use that default. Else you"
          " can type NONE to not limit the persist dir, or any number you like"
          " (a GiB value, so type 0.5 to allow ~500 MiB)." % default)
    value = raw_input('persist-size-limit: ')
    if value == "NONE":
        return 0.0
    elif value:
        return float(value)
    return default


def _get_bits(default):
    print("You are using a 64-bit system, so mozregression will by default"
          " use the 64-bit build files. If you want to change that to"
          " 32-bit by default, type 32 here.")
    while 1:
        value = raw_input('bits: ')
        if value in ('', '32', '64'):
            break
    if not value:
        return default
    return value


CONF_HELP = """\
# ------ mozregression configuration file ------

# Most of the command line options can be used in here.
# Just remove the -- from the long option names, e.g.

# bits = 32
# persist-size-limit = 15.0


"""


def write_conf(conf_path):
    conf_dir = os.path.dirname(conf_path)
    if not os.path.isdir(conf_dir):
        os.makedirs(conf_dir)

    config = ConfigObj(conf_path)
    if not config.initial_comment:
        config.initial_comment = CONF_HELP.splitlines()

    def _set_option(optname, getfunc, default):
        print
        if optname not in config:
            value = getfunc(default)
            if value is not None:
                config[optname] = value
            else:
                value = default
        else:
            print '%s already defined.' % optname
            value = config[optname]
        name = colorize("{fGREEN}%s{sRESET_ALL}" % optname)
        print "%s: %s" % (name, value)

    _set_option('persist', _get_persist_dir,
                os.path.join(conf_dir, "persist"))

    _set_option('persist-size-limit', _get_persist_size_limit, 20.0)

    if mozinfo.os != 'mac' and mozinfo.bits == 64:
        _set_option('bits', _get_bits, 64)

    config.write()

    print
    print colorize('Config file {sBRIGHT}%s{sRESET_ALL} written.' % conf_path)
    print("Note you can edit it manually, and there are other options you can"
          " configure. See %s." % CONFIG_FILE_HELP_URL)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
This module parses and checks the command line with :func:`cli` and return a
:class:`Configuration` object that hold information for running the
application.

:func:`cli` is intended to be the only public interface of this module.
"""

import os
import mozinfo
import datetime
import mozprofile
import re

from argparse import ArgumentParser, Action, SUPPRESS
from mozlog.structuredlog import get_default_logger

from mozregression import __version__
from mozregression.branches import get_name
from mozregression.dates import to_datetime, parse_date, is_date_or_datetime
from mozregression.config import get_defaults, DEFAULT_CONF_FNAME, write_conf
from mozregression.tc_authenticate import tc_authenticate
from mozregression.fetch_configs import REGISTRY as FC_REGISTRY, create_config
from mozregression.errors import (MozRegressionError, DateFormatError,
                                  UnavailableRelease)
from mozregression.releases import (formatted_valid_release_dates,
                                    date_of_release, tag_of_release,
                                    tag_of_beta)
from mozregression.log import init_logger, colorize


class _StopAction(Action):
    def __init__(self, option_strings, dest=SUPPRESS, default=SUPPRESS,
                 help=None):
        super(_StopAction, self).__init__(option_strings=option_strings,
                                          dest=dest, default=default,
                                          nargs=0, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        raise NotImplementedError


class ListReleasesAction(_StopAction):
    def __call__(self, parser, namespace, values, option_string=None):
        print(formatted_valid_release_dates())
        parser.exit()


class WriteConfigAction(_StopAction):
    def __call__(self, parser, namespace, values, option_string=None):
        write_conf(DEFAULT_CONF_FNAME)
        parser.exit()


class ListBuildTypesAction(_StopAction):
    def __call__(self, parser, namespace, values, option_string=None):
        for name in FC_REGISTRY.names():
            print("%s:" % name)
            klass = FC_REGISTRY.get(name)
            for btype in klass.BUILD_TYPES:
                print("  %s" % btype)
        parser.exit()


def parse_args(argv=None, defaults=None):
    """
    Parse command line options.
    """
    parser = create_parser(defaults=defaults)
    return parser.parse_args(argv)


def create_parser(defaults):
    """
    Create the mozregression command line parser (ArgumentParser instance).
    """
    usage = ("\n"
             " %(prog)s [OPTIONS]"
             " [--bad DATE|BUILDID|RELEASE|CHANGESET]"
             " [--good DATE|BUILDID|RELEASE|CHANGESET]"
             "\n"
             " %(prog)s [OPTIONS] --launch DATE|BUILDID|RELEASE|CHANGESET"
             "\n"
             " %(prog)s --list-build-types"
             "\n"
             " %(prog)s --list-releases"
             "\n"
             " %(prog)s --write-conf")

    parser = ArgumentParser(usage=usage)
    parser.add_argument("--version", action="version", version=__version__,
                        help=("print the mozregression version number and"
                              " exits."))

    parser.add_argument("-b", "--bad",
                        metavar="DATE|BUILDID|RELEASE|CHANGESET",
                        help=("first known bad build, default is today."
                              " It can be a date (YYYY-MM-DD), a build id,"
                              " a release number or a changeset. If it is"
                              " a changeset, the default branch will be the"
                              " integration branch of the application"
                              " (e.g. mozilla-inbound for firefox), else"
                              " the default release branch for the application"
                              " will be used as the default (e.g"
                              " mozilla-central for firefox)."))

    parser.add_argument("-g", "--good",
                        metavar="DATE|BUILDID|RELEASE|CHANGESET",
                        help=("last known good build. Same possible values"
                              " as the --bad option."))

    parser.add_argument("--list-releases",
                        action=ListReleasesAction,
                        help="list all known releases and exit")

    parser.add_argument("-B", "--build-type",
                        default=defaults["build-type"],
                        help=("Build type to use, e.g. opt, debug. "
                              "See --list-build-types for available values. "
                              "Defaults to shippable for desktop Fx, opt for "
                              "everything else."))

    parser.add_argument("--list-build-types", action=ListBuildTypesAction,
                        help="List available build types combinations.")

    parser.add_argument("--find-fix", action="store_true",
                        help="Search for a bug fix instead of a regression.")

    parser.add_argument("-e", "--addon",
                        dest="addons",
                        action='append',
                        default=[],
                        metavar="PATH1",
                        help="addon to install; repeat for multiple addons.")

    parser.add_argument("-p", "--profile",
                        default=defaults["profile"],
                        metavar="PATH",
                        help="profile to use with nightlies.")

    parser.add_argument('--adb-profile-dir',
                        dest="adb_profile_dir",
                        default=defaults["adb-profile-dir"],
                        help=("Path to use on android devices for storing"
                              " the profile. Generally you should not need"
                              " to specify this, and an appropriate path"
                              " will be used. Specifying this to a value,"
                              " e.g. '/sdcard/tests' will forcibly try to create"
                              " the profile inside that folder."))

    parser.add_argument('--profile-persistence',
                        choices=('clone', 'clone-first', 'reuse'),
                        default=defaults["profile-persistence"],
                        help=("Persistence of the used profile. Before"
                              " each tested build, a profile is used. If"
                              " the value of this option is 'clone', each"
                              " test uses a fresh clone. If the value is"
                              " 'clone-first', the profile is cloned once"
                              " then reused for all builds during the "
                              " bisection. If the value is 'reuse', the given"
                              " profile is directly used. Note that the"
                              " profile might be modified by mozregression."
                              " Defaults to %(default)s."))

    parser.add_argument("-a", "--arg",
                        dest="cmdargs",
                        action='append',
                        default=[],
                        metavar="ARG1",
                        help=("a command-line argument to pass to the"
                              " application; repeat for multiple arguments."
                              " Use --arg='-option' to pass in options"
                              " starting with `-`."))

    parser.add_argument('--pref', nargs='*', dest='prefs',
                        help=(" A preference to set. Must be a key-value pair"
                              " separated by a ':'. Note that if your"
                              " preference is of type float, you should"
                              " pass it as a string, e.g.:"
                              " --pref \"layers.low-precision-opacity:'0.0'\""
                              ))

    parser.add_argument('--preferences', nargs="*", dest='prefs_files',
                        help=("read preferences from a JSON or INI file. For"
                              " INI, use 'file.ini:section' to specify a"
                              " particular section."))

    parser.add_argument("-n", "--app",
                        choices=FC_REGISTRY.names(),
                        default=defaults["app"],
                        help="application name. Default: %(default)s.")

    parser.add_argument("--repo",
                        metavar="[mozilla-inbound|autoland|mozilla-beta...]",
                        default=defaults["repo"],
                        help="repository name used for the bisection.")

    parser.add_argument("--bits",
                        choices=("32", "64"),
                        default=defaults["bits"],
                        help=("force 32 or 64 bit version (only applies to"
                              " x86_64 boxes). Default: %s bits."
                              % defaults["bits"] or mozinfo.bits))

    parser.add_argument("-c", "--command",
                        help=("Test command to evaluate builds automatically."
                              " A return code of 0 will evaluate the build as"
                              " good, and any other value as bad."
                              " Variables like {binary} can be used, which"
                              " will be replaced with their value as retrieved"
                              " by the actual build."
                              ))

    parser.add_argument("--persist",
                        default=defaults["persist"],
                        help=("the directory in which downloaded files are"
                              " to persist. Defaults to %(default)r."))

    parser.add_argument('--persist-size-limit', type=float,
                        default=defaults['persist-size-limit'],
                        help=("Size limit for the persist directory in"
                              " gigabytes (GiB). When the limit is reached,"
                              " old builds are removed. 0 means no limit. Note"
                              " that at least 5 build files are kept,"
                              " regardless of this value."
                              " Defaults to %(default)s."))

    parser.add_argument('--http-timeout', type=float,
                        default=float(defaults['http-timeout']),
                        help=("Timeout in seconds to abort requests when there"
                              " is no activity from the server. Default to"
                              " %(default)s seconds - increase this if you"
                              " are under a really slow network."))

    parser.add_argument('--no-background-dl', action='store_false',
                        dest="background_dl",
                        default=(defaults['no-background-dl'].lower()
                                 not in ('1', 'yes', 'true')),
                        help=("Do not download next builds in the background"
                              " while evaluating the current build."))

    parser.add_argument('--background-dl-policy', choices=('cancel', 'keep'),
                        default=defaults['background-dl-policy'],
                        help=('Policy to use for background downloads.'
                              ' Possible values are "cancel" to cancel all'
                              ' pending background downloads or "keep" to keep'
                              ' downloading them when persist mode is enabled.'
                              ' The default is %(default)s.'))

    parser.add_argument('--approx-policy', choices=('auto', 'none'),
                        default=defaults['approx-policy'],
                        help=("Policy to reuse approximate persistent builds"
                              " instead of downloading the accurate ones."
                              " When auto, mozregression will try its best to"
                              " reuse the files, meaning that for 7 days of"
                              " bisection range it will try to reuse a build"
                              " which date approximates the build to download"
                              " by one day (before or after). Use none to"
                              " disable this behavior."
                              " Defaults to %(default)s."))

    parser.add_argument('--launch',
                        metavar="DATE|BUILDID|RELEASE|CHANGESET",
                        help=("Launch only one specific build. Same possible"
                              " values as the --bad option."))

    parser.add_argument('-P', '--process-output', choices=('none', 'stdout'),
                        default=defaults['process-output'],
                        help=("Manage process output logging. Set to stdout by"
                              " default when the build type is not 'opt'."))

    parser.add_argument('-M', '--mode', choices=('classic', 'no-first-check'),
                        default=defaults['mode'],
                        help=("bisection mode. 'classic' will check for the"
                              " first good and bad builds to really be good"
                              " and bad, and 'no-first-check' won't. Defaults"
                              " to %(default)s."))

    parser.add_argument('--archive-base-url',
                        default=defaults['archive-base-url'],
                        help=("Base url used to find the archived builds."
                              " Defaults to %(default)s"))

    parser.add_argument('--write-config',
                        action=WriteConfigAction,
                        help="Helps to write the configuration file.")

    parser.add_argument('--debug', '-d', action='store_true',
                        help='Show the debug output.')

    return parser


def parse_bits(option_bits):
    """
    Returns the correctly typed bits.
    """
    if option_bits == "32":
        return 32
    else:
        # if 64 bits is passed on a 32 bit system, it won't be honored
        return mozinfo.bits


def preferences(prefs_files, prefs_args, logger):
    """
    profile preferences
    """
    # object that will hold the preferences
    prefs = mozprofile.prefs.Preferences()

    # add preferences files
    if prefs_files:
        for prefs_file in prefs_files:
            prefs.add_file(prefs_file)

    separator = ':'
    cli_prefs = []
    if prefs_args:
        for pref in prefs_args:
            if separator not in pref:
                if logger:
                    if '=' in pref:
                        logger.warning('Pref %s has an "=", did you mean to use ":"?' % pref)
                    logger.info('Dropping pref %s for missing separator ":"' % pref)
                continue
            cli_prefs.append(pref.split(separator, 1))

    # string preferences
    prefs.add(cli_prefs, cast=True)

    return prefs()


def get_default_date_range(fetch_config):
    """
    Compute the default date range (first, last) to bisect.
    """
    last_date = datetime.date.today()
    if fetch_config.app_name == 'jsshell':
        if fetch_config.os == "win" and fetch_config.bits == 64:
            first_date = datetime.date(2014, 5, 27)
        elif fetch_config.os == "linux" and "asan" in fetch_config.build_type:
            first_date = datetime.date(2013, 9, 1)
        else:
            first_date = datetime.date(2012, 4, 18)
    elif fetch_config.os == 'win' and fetch_config.bits == 64:
        # first firefox build date for win64 is 2010-05-28
        first_date = datetime.date(2010, 5, 28)
    else:
        first_date = datetime.date(2009, 1, 1)

    return first_date, last_date


class Configuration(object):
    """
    Holds the configuration extracted from the command line.

    This is usually instantiated by calling :func:`cli`.

    The constructor only initializes the `logger`.

    The configuration should not be used (except for the logger attribute)
    until :meth:`validate` is called.

    :attr logger: the mozlog logger, created using the command line options
    :attr options: the raw command line options
    :attr action: the action that the user want to do. This is a string
                  ("bisect_inbounds" or "bisect_nightlies")
    :attr fetch_config: the fetch_config instance, required to find
                        information about a build
    """
    def __init__(self, options):
        self.options = options
        self.logger = init_logger(debug=options.debug)
        # allow to filter process output based on the user option
        if options.process_output is None:
            # process_output not user defined
            log_process_output = options.build_type != ''
        else:
            log_process_output = options.process_output == 'stdout'
        get_default_logger("process").component_filter = \
            lambda data: data if log_process_output else None

        # filter some mozversion log lines
        re_ignore_mozversion_line = re.compile(
            r"^(platform_.+|application_vendor|application_remotingname"
            r"|application_id|application_display_name): .+"
        )
        get_default_logger("mozversion").component_filter = lambda data: (
            None if re_ignore_mozversion_line.match(data['message']) else data
        )

        self.action = None
        self.fetch_config = None

    def _convert_to_bisect_arg(self, value):
        """
        Transform a string value to a date or datetime if it looks like it.
        """
        try:
            value = parse_date(value)
        except DateFormatError:
            try:
                repo = self.options.repo
                if (get_name(repo) == 'mozilla-release' or
                        (not repo and re.match(r'^\d+\.\d\.\d$', value))):
                    new_value = tag_of_release(value)
                    if not repo:
                        self.logger.info("Assuming repo mozilla-release")
                        self.fetch_config.set_repo('mozilla-release')
                    self.logger.info("Using tag %s for release %s"
                                     % (new_value, value))
                    value = new_value
                elif (get_name(repo) == 'mozilla-beta' or
                      (not repo and re.match(r'^\d+\.0b\d+$', value))):
                    new_value = tag_of_beta(value)
                    if not repo:
                        self.logger.info("Assuming repo mozilla-beta")
                        self.fetch_config.set_repo('mozilla-beta')
                    self.logger.info("Using tag %s for release %s"
                                     % (new_value, value))
                    value = new_value
                else:
                    new_value = parse_date(date_of_release(value))
                    self.logger.info("Using date %s for release %s"
                                     % (new_value, value))
                    value = new_value
            except UnavailableRelease:
                self.logger.info("%s is not a release, assuming it's a hash..." % value)
        return value

    def validate(self):
        """
        Validate the options, define the `action` and `fetch_config` that
        should be used to run the application.
        """
        options = self.options

        user_defined_bits = options.bits is not None
        options.bits = parse_bits(options.bits or mozinfo.bits)
        fetch_config = create_config(options.app, mozinfo.os, options.bits,
                                     mozinfo.processor)
        if options.build_type:
            try:
                fetch_config.set_build_type(options.build_type)
            except MozRegressionError as msg:
                self.logger.warning(
                    "%s (Defaulting to %r)" % (msg, fetch_config.build_type)
                )
        self.fetch_config = fetch_config

        fetch_config.set_repo(options.repo)
        if fetch_config.is_nightly():
            fetch_config.set_base_url(options.archive_base_url)

        if not user_defined_bits and \
                options.bits == 64 and \
                mozinfo.os == 'win' and \
                32 in fetch_config.available_bits():
            # inform users on windows that we are using 64 bit builds.
            self.logger.info("bits option not specified, using 64-bit builds.")

        if options.bits == 32 and mozinfo.os == 'mac':
            self.logger.info("only 64-bit builds available for mac, using "
                             "64-bit builds")

        if fetch_config.is_inbound() and fetch_config.tk_needs_auth():
            creds = tc_authenticate(self.logger)
            fetch_config.set_tk_credentials(creds)

        # set action for just use changset or data to bisect
        if options.launch:
            options.launch = self._convert_to_bisect_arg(options.launch)
            self.action = "launch_inbound"
            if is_date_or_datetime(options.launch) and \
                    not fetch_config.should_use_taskcluster():
                self.action = "launch_nightlies"
        else:
            # define good/bad default values if required
            default_good_date, default_bad_date = \
                get_default_date_range(fetch_config)
            if options.find_fix:
                default_bad_date, default_good_date = \
                    default_good_date, default_bad_date
            if not options.bad:
                options.bad = default_bad_date
                self.logger.info("No 'bad' option specified, using %s"
                                 % options.bad)
            else:
                options.bad = self._convert_to_bisect_arg(options.bad)
            if not options.good:
                options.good = default_good_date
                self.logger.info("No 'good' option specified, using %s"
                                 % options.good)
            else:
                options.good = self._convert_to_bisect_arg(options.good)

            self.action = "bisect_inbounds"
            if is_date_or_datetime(options.good) and \
                    is_date_or_datetime(options.bad):
                if not options.find_fix and \
                        to_datetime(options.good) > to_datetime(options.bad):
                    raise MozRegressionError(
                        ("Good date %s is later than bad date %s."
                         " Maybe you wanted to use the --find-fix"
                         " flag?") % (options.good, options.bad))
                elif options.find_fix and \
                        to_datetime(options.good) < to_datetime(options.bad):
                    raise MozRegressionError(
                        ("Bad date %s is later than good date %s."
                         " You should not use the --find-fix flag"
                         " in this case...") % (options.bad, options.good))
                if not fetch_config.should_use_taskcluster():
                    self.action = "bisect_nightlies"
        if self.action in ('launch_inbound', 'bisect_inbounds')\
                and not fetch_config.is_inbound():
            raise MozRegressionError('Unable to bisect inbound for `%s`'
                                     % fetch_config.app_name)
        options.preferences = preferences(options.prefs_files, options.prefs, self.logger)
        # convert GiB to bytes.
        options.persist_size_limit = \
            int(abs(float(options.persist_size_limit)) * 1073741824)


def cli(argv=None, conf_file=DEFAULT_CONF_FNAME, namespace=None):
    """
    parse cli args basically and returns a :class:`Configuration`.

    if namespace is given, it will be used as a arg parsing result, so no
    arg parsing will be done.
    """
    if namespace:
        options = namespace
    else:
        defaults = get_defaults(conf_file)
        options = parse_args(argv=argv, defaults=defaults)
        if not options.cmdargs:
            # we don't set the cmdargs default to be that from the
            # configuration file, because then any new arguments
            # will be appended: https://bugs.python.org/issue16399
            options.cmdargs = defaults['cmdargs']
    if conf_file and not os.path.isfile(conf_file):
        print '*' * 10
        print colorize("You should use a config file. Please use the " +
                       '{sBRIGHT}--write-config{sRESET_ALL}' +
                       " command line flag to help you create one.")
        print '*' * 10
        print
    return Configuration(options)



"""
Date utilities functions.
"""

import re
import datetime
import calendar

from mozregression.errors import DateFormatError


def parse_date(date_string):
    """
    Returns a date or a datetime from a string.
    """
    if len(date_string) == 14 and date_string.isdigit():
        # probably a build id - transform that in a datetime
        try:
            return datetime.datetime.strptime(date_string, "%Y%m%d%H%M%S")
        except ValueError:
            raise DateFormatError(date_string, "Not a valid build id: `%s`")
    regex = re.compile(r'(\d{4})\-(\d{1,2})\-(\d{1,2})')
    matched = regex.match(date_string)
    if not matched:
        raise DateFormatError(date_string)
    return datetime.date(int(matched.group(1)),
                         int(matched.group(2)),
                         int(matched.group(3)))


def to_datetime(date):
    """
    transform a date to a datetime instance
    If the parameter is not a date, it is returned without modification
    """
    if isinstance(date, datetime.date):
        return datetime.datetime.combine(date, datetime.time())
    return date


def to_date(date_time):
    """
    transform a datetime to a date instance
    If the parameter is not a datetime, it is returned without modification
    """
    if isinstance(date_time, datetime.datetime):
        return date_time.date()
    return date_time


def is_date_or_datetime(obj):
    return isinstance(obj, (datetime.date, datetime.datetime))


def to_utc_timestamp(date_time):
    return calendar.timegm(date_time.timetuple())



"""
Logging and outputting configuration and utilities.
"""

import sys
import time
import mozinfo

from colorama import Fore, Style, Back
from mozlog.structuredlog import set_default_logger, StructuredLogger
from mozlog.handlers import StreamHandler, LogLevelFilter

ALLOW_COLOR = sys.stdout.isatty()


def _format_seconds(total):
    """Format number of seconds to MM:SS.DD form."""
    minutes, seconds = divmod(total, 60)
    return '%2d:%05.2f' % (minutes, seconds)


def init_logger(debug=True, allow_color=ALLOW_COLOR, output=None):
    """
    Initialize the mozlog logger. Must be called once before using logs.
    """
    # late binding of sys.stdout is required for windows color to work
    output = output or sys.stdout
    start = time.time() * 1000
    level_color = {
        'WARNING': Fore.MAGENTA + Style.BRIGHT,
        'CRITICAL': Fore.RED + Style.BRIGHT,
        'ERROR': Fore.RED + Style.BRIGHT,
        'DEBUG': Fore.CYAN + Style.BRIGHT,
        'INFO': Style.BRIGHT,
    }
    time_color = Fore.BLUE
    if mozinfo.os == "win":
        time_color += Style.BRIGHT  # this is unreadable on windows without it

    def format_log(data):
        level = data['level']
        elapsed = _format_seconds((data['time'] - start) / 1000)
        if allow_color:
            elapsed = time_color + elapsed + Style.RESET_ALL
            if level in level_color:
                level = level_color[level] + level + Style.RESET_ALL
        msg = data['message']
        if 'stack' in data:
            msg += "\n%s" % data['stack']
        return "%s %s: %s\n" % (elapsed, level, msg)

    logger = StructuredLogger("mozregression")
    handler = LogLevelFilter(StreamHandler(output, format_log),
                             'debug' if debug else 'info')
    logger.add_handler(handler)

    set_default_logger(logger)
    return logger


COLORS = {}
NO_COLORS = {}

for prefix, st in (('b', Back), ('s', Style), ('f', Fore)):
    for name, value in st.__dict__.iteritems():
        COLORS[prefix + name] = value
        NO_COLORS[prefix + name] = ''


def colorize(text, allow_color=ALLOW_COLOR):
    """
    *colorize* text to be displayed on terminal.

    You can pass a string with key parameters to be formatted. you can use
    every name available from colorama.{Back,Style,Fore}, with corresponding
    prefixes, followed by the property you want to use. Prefixes are:

    - Back: "b"
    - Style: "s"
    - Fore: "f"

    Example::

    >> colorize("{fRED}hellow{sRESET_ALL}")

    Will colorize the text on the screen if allow_color is True (equivalent to
    Fore.RED + "hello" + Style.RESET_ALL).
    If allow_color is False, no color special char will be added, thus the
    returned text will be "hello".
    """
    data = COLORS if allow_color else NO_COLORS
    return text.format(**data)



"""
This module defines the configuration needed for nightly and inbound
fetching for each application. This configuration is a base block
for everything done in mozregression since it holds information
about how to get information about builds for a given application.

The public entry point in there is :func:`create_config`, which
creates an returns a fetch configuration. the configuration will
be an instance of :class:`CommonConfig`, possibly using the mixins
:class:`NightlyConfigMixin` and/or :class:`InboundConfigMixin`.
<
Example to create a configuration for firefox on linux 64: ::

  fetch_config = create_config('firefox', 'linux', 64)

You can also use the variable *REGISTRY* defined in this module to get a
list of application names that can be used to build a configuration. This is
an instance of :class:`ClassRegistry`. Example: ::

  print REGISTRY.names()
"""
import re
import datetime

from mozregression.class_registry import ClassRegistry
from mozregression import errors, branches
from mozregression.dates import to_utc_timestamp
from mozregression.config import ARCHIVE_BASE_URL
from abc import ABCMeta, abstractmethod


# switch from fennec api-11 to api-15 on taskcluster
# appeared on this date for m-c.
TIMESTAMP_FENNEC_API_15 = to_utc_timestamp(
    datetime.datetime(2016, 1, 29, 0, 30, 13)
)

# switch from fennec api-15 to api-16 on taskcluster
# appeared on this date for m-c.
TIMESTAMP_FENNEC_API_16 = to_utc_timestamp(
    datetime.datetime(2017, 8, 29, 18, 28, 36)
)


def get_build_regex(name, os, bits, processor, psuffix='', with_ext=True):
    """
    Returns a string regexp that can match a build filename.

    :param name: must be the beginning of the filename to match
    :param os: the os, as returned by mozinfo.os
    :param bits: the bits information of the build. Either 32 or 64.
    :param processor: the architecture of the build. Only one that alters
                      results is aarch64.
    :param psuffix: optional suffix before the extension
    :param with_ext: if True, the build extension will be appended (either
                     .zip, .tar.bz2 or .dmg depending on the os).
    """
    if os == "win":
        if bits == 64:
            if processor == "aarch64":
                suffix = r".*win64-aarch64"
            else:
                suffix = r".*win64(-x86_64)?"
            ext = r"\.zip"
        else:
            suffix, ext = r".*win32", r"\.zip"
    elif os == "linux":
        if bits == 64:
            suffix, ext = r".*linux-x86_64", r"\.tar.bz2"
        else:
            suffix, ext = r".*linux-i686", r"\.tar.bz2"
    elif os == "mac":
        suffix, ext = r".*mac.*", r"\.dmg"
    else:
        raise errors.MozRegressionError(
            "mozregression supports linux, mac and windows but your"
            " os is reported as '%s'." % os
        )

    # New taskcluster builds now just name the binary archive 'target', so
    # that is added as one possibility in the regex.
    regex = '(target|%s%s%s)' % (name, suffix, psuffix)
    if with_ext:
        return '%s%s' % (regex, ext)
    else:
        return regex


class CommonConfig(object):
    """
    Define the configuration for both nightly and inbound fetching.
    """
    BUILD_TYPES = ('opt',)  # only opt allowed by default
    BUILD_TYPE_FALLBACKS = {}
    app_name = None

    def __init__(self, os, bits, processor):
        self.os = os
        self.bits = bits
        self.processor = processor
        self.repo = None
        self.set_build_type('opt')
        self._used_build_index = 0

    @property
    def build_type(self):
        """
        Returns the currently selected build type, which can change if there
        are fallbacks specified.
        """
        return self.build_types[self._used_build_index]

    def _inc_used_build(self):
        """
        Increments the index into the build_types indicating the currently
        selected build type.
        """
        self._used_build_index = (
            # Need to be careful not to overflow the list
            (self._used_build_index + 1) % len(self.build_types)
        )

    def build_regex(self):
        """
        Returns a string regex that can match a build file on the servers.
        """
        return get_build_regex(self.app_name, self.os, self.bits,
                               self.processor) + '$'

    def build_info_regex(self):
        """
        Returns a string regex that can match a build info file (txt)
        on the servers.
        """
        return get_build_regex(self.app_name, self.os, self.bits,
                               self.processor, with_ext=False) + r'\.txt$'

    def is_nightly(self):
        """
        Returns True if the configuration can be used for nightly fetching.
        """
        return isinstance(self, NightlyConfigMixin)

    def is_inbound(self):
        """
        Returns True if the configuration can be used for inbound fetching.
        """
        return isinstance(self, InboundConfigMixin)

    def available_bits(self):
        """
        Returns the no. of bits of the OS for which the application should
        run.
        """
        return (32, 64)

    def available_build_types(self):
        res = []
        for available in self.BUILD_TYPES:
            match = re.match("(.+)\[(.+)\]", available)
            if match:
                suffix = ('-aarch64' if self.processor == 'aarch64' and
                          self.bits == 64 else '')
                available = match.group(1)
                platforms = match.group(2)
                if '{}{}{}'.format(
                    self.os, self.bits, suffix
                ) not in platforms.split(','):
                    available = None
            if available:
                res.append(available)
        return res

    def set_build_type(self, build_type):
        """
        Define the build type (opt, debug, asan...).

        :raises: MozRegressionError on error.
        """
        if build_type in self.available_build_types():
            fallbacks = self.BUILD_TYPE_FALLBACKS.get(build_type)
            self.build_types = (
                (build_type,) + fallbacks if fallbacks else (build_type,)
            )
            return
        raise errors.MozRegressionError(
            "Unable to find a suitable build type %r." % str(build_type)
        )

    def set_repo(self, repo):
        """
        Allow to define the repo name.

        If not set or set to None, default repos would be used (see
        :meth:`get_nightly_repo` and :attr:`inbound_branch`)
        """
        self.repo = branches.get_name(repo) if repo else None

    def should_use_taskcluster(self):
        """
        Returns True if taskcluster should be used as the bisection method.

        Note that this method relies on the repo and build type defined.
        """
        return (branches.get_category(self.repo) in
                ('integration', 'try', 'releases') or
                # we can find the asan builds (firefox and jsshell) in
                # archives.m.o
                self.build_type not in ('opt', 'asan', 'shippable'))


class NightlyConfigMixin:
    """
    Define the nightly-related required configuration to find nightly builds.

    A nightly build url is divided in 2 parts here:

    1. the base part as returned by :meth:`get_nighly_base_url`
    2. the final part, which can be found using :meth:`get_nighly_repo_regex`

    The final part contains a repo name, which is returned by
    :meth:`get_nightly_repo`.

    Note that subclasses must implement :meth:`_get_nightly_repo` to
    provide a default value.
    """
    archive_base_url = ARCHIVE_BASE_URL
    __metaclass__ = ABCMeta
    nightly_base_repo_name = "firefox"
    nightly_repo = None

    def set_base_url(self, url):
        self.archive_base_url = url.rstrip('/')

    def get_nighly_base_url(self, date):
        """
        Returns the base part of the nightly build url for a given date.
        """
        return "%s/%s/nightly/%04d/%02d/" % (self.archive_base_url,
                                             self.nightly_base_repo_name,
                                             date.year,
                                             date.month)

    def get_nightly_repo(self, date):
        """
        Returns the repo name for a given date.
        """
        if isinstance(date, datetime.datetime):
            date = date.date()
        return self.repo or self._get_nightly_repo(date)

    @abstractmethod
    def _get_nightly_repo(self, date):
        """
        Returns a default repo name for a given date.
        """
        raise NotImplementedError

    def get_nightly_repo_regex(self, date):
        """
        Returns a string regex that can match the last folder name for a given
        date.
        """
        return self._get_nightly_repo_regex(date, self.get_nightly_repo(date))

    def _get_nightly_repo_regex(self, date, repo):
        if isinstance(date, datetime.datetime):
            return (r'^%04d-%02d-%02d-%02d-%02d-%02d-%s/$'
                    % (date.year, date.month, date.day, date.hour,
                       date.minute, date.second, repo))
        return (r'^%04d-%02d-%02d-[\d-]+%s/$'
                % (date.year, date.month, date.day, repo))

    def can_go_inbound(self):
        """
        Indicate if we can bisect inbound from this nightly config.
        """
        return self.is_inbound()


class FireFoxNightlyConfigMixin(NightlyConfigMixin):
    def _get_nightly_repo(self, date):
        if date < datetime.date(2008, 6, 17):
            return "trunk"
        else:
            return "mozilla-central"


class ThunderbirdNightlyConfigMixin(NightlyConfigMixin):
    nightly_base_repo_name = 'thunderbird'

    def _get_nightly_repo(self, date):
        # sneaking this in here
        if self.os == "win" and date < datetime.date(2010, 3, 18):
            # no .zip package for Windows, can't use the installer
            raise errors.WinTooOldBuildError()

        if date < datetime.date(2008, 7, 26):
            return "trunk"
        elif date < datetime.date(2009, 1, 9):
            return "comm-central"
        elif date < datetime.date(2010, 8, 21):
            return "comm-central-trunk"
        else:
            return "comm-central"


class FennecNightlyConfigMixin(NightlyConfigMixin):
    nightly_base_repo_name = "mobile"

    def _get_nightly_repo(self, date):
        return 'mozilla-central'

    def get_nightly_repo_regex(self, date):
        repo = self.get_nightly_repo(date)
        if repo in ('mozilla-central',):
            if date < datetime.date(2014, 12, 6):
                repo += "-android"
            elif date < datetime.date(2014, 12, 13):
                repo += "-android-api-10"
            elif date < datetime.date(2016, 1, 29):
                repo += "-android-api-11"
            elif date < datetime.date(2017, 8, 30):
                repo += "-android-api-15"
            else:
                repo += "-android-api-16"
        return self._get_nightly_repo_regex(date, repo)


class InboundConfigMixin:
    """
    Define the inbound-related required configuration.
    """
    __metaclass__ = ABCMeta
    default_inbound_branch = 'mozilla-inbound'
    _tk_credentials = None

    @property
    def inbound_branch(self):
        return self.repo or self.default_inbound_branch

    def tk_inbound_route(self, push):
        """
        Returns the first taskcluster route for a specific changeset
        """
        return next(self.tk_inbound_routes(push))

    @abstractmethod
    def tk_inbound_routes(self, push):
        """
        Returns a generator of taskcluster routes for a specific changeset.
        """
        raise NotImplementedError

    def inbound_persist_part(self):
        """
        Allow to add a part in the generated persist file name to distinguish
        builds. Returns an empty string by default, or 'debug' if build type
        is debug.
        """
        return self.build_type if self.build_type != 'opt' else ''

    def tk_needs_auth(self):
        """
        Returns True if we need taskcluster credentials
        """
        return False

    def set_tk_credentials(self, creds):
        """
        Define the credentials required to download private builds on
        TaskCluster.
        """
        self._tk_credentials = creds

    def tk_options(self):
        """
        Returns the takcluster options, including the credentials required to
        download private artifacts.
        """
        tk_options = {'rootUrl': 'https://taskcluster.net'}
        if self.tk_needs_auth():
            tk_options.update({'credentials': self._tk_credentials})
        return tk_options


def _common_tk_part(inbound_conf):
    # private method to avoid copy/paste for building taskcluster route part.
    if inbound_conf.os == 'linux':
        part = 'linux'
        if inbound_conf.bits == 64:
            part += str(inbound_conf.bits)
    elif inbound_conf.os == 'mac':
        part = 'macosx64'
    else:
        # windows
        part = '{}{}'.format(inbound_conf.os, inbound_conf.bits)
        if inbound_conf.processor == 'aarch64' and inbound_conf.bits == 64:
            part += '-aarch64'
    return part


class FirefoxInboundConfigMixin(InboundConfigMixin):
    def tk_inbound_routes(self, push):
        for build_type in self.build_types:
            yield 'gecko.v2.{}{}.revision.{}.firefox.{}-{}'.format(
                self.inbound_branch,
                '.shippable' if build_type == 'shippable' else '',
                push.changeset,
                _common_tk_part(self),
                'opt' if build_type == 'shippable' else build_type
            )
            self._inc_used_build()
        return


class FennecInboundConfigMixin(InboundConfigMixin):
    tk_name = 'android-api-11'

    def tk_inbound_routes(self, push):
        tk_name = self.tk_name
        if tk_name == 'android-api-11':
            if push.timestamp >= TIMESTAMP_FENNEC_API_16:
                tk_name = 'android-api-16'
            elif push.timestamp >= TIMESTAMP_FENNEC_API_15:
                tk_name = 'android-api-15'
        for build_type in self.build_types:
            yield 'gecko.v2.{}.revision.{}.mobile.{}-{}'.format(
                self.inbound_branch, push.changeset, tk_name,
                build_type
            )
            self._inc_used_build()
            return


class ThunderbirdInboundConfigMixin(InboundConfigMixin):
    default_inbound_branch = 'comm-central'

    def tk_inbound_routes(self, push):
        for build_type in self.build_types:
            yield 'comm.v2.{}.revision.{}.thunderbird.{}-{}'.format(
                self.inbound_branch, push.changeset, _common_tk_part(self),
                build_type
            )
            self._inc_used_build()
        return

# ------------ full config implementations ------------


REGISTRY = ClassRegistry('app_name')


def create_config(name, os, bits, processor):
    """
    Create and returns a configuration for the given name.

    :param name: application name, such as 'firefox'
    :param os: os name, e.g 'linux', 'win' or 'mac'
    :param bits: the bit of the os as an int, e.g 32 or 64. Can be None
                 if the bits do not make sense (e.g. fennec)
    :param processor: processor family, e.g 'x86', 'x86_86', 'ppc', 'ppc64' or
                      'aarch64'
    """
    return REGISTRY.get(name)(os, bits, processor)


@REGISTRY.register('firefox')
class FirefoxConfig(CommonConfig,
                    FireFoxNightlyConfigMixin,
                    FirefoxInboundConfigMixin):
    BUILD_TYPES = ('shippable', 'opt', 'pgo[linux32,linux64,win32,win64]',
                   'debug', 'asan[linux64]', 'asan-debug[linux64]')
    BUILD_TYPE_FALLBACKS = {
        'shippable': ('opt', 'pgo'),
        'opt': ('shippable', 'pgo'),
    }

    def __init__(self, os, bits, processor):
        super(FirefoxConfig, self).__init__(os, bits, processor)
        self.set_build_type('shippable')

    def build_regex(self):
        return get_build_regex(
            self.app_name, self.os, self.bits, self.processor,
            psuffix='-asan' if 'asan' in self.build_type else ''
        ) + '$'


@REGISTRY.register('thunderbird')
class ThunderbirdConfig(CommonConfig,
                        ThunderbirdNightlyConfigMixin,
                        ThunderbirdInboundConfigMixin):
    pass


@REGISTRY.register('fennec')
class FennecConfig(CommonConfig,
                   FennecNightlyConfigMixin,
                   FennecInboundConfigMixin):
    BUILD_TYPES = ('opt', 'debug')

    def build_regex(self):
        return r'(target|fennec-.*)\.apk'

    def build_info_regex(self):
        return r'(target|fennec-.*)\.txt'

    def available_bits(self):
        return ()


@REGISTRY.register('gve')
class GeckoViewExampleConfig(CommonConfig,
                             FennecNightlyConfigMixin,
                             FennecInboundConfigMixin):
    BUILD_TYPES = ('opt', 'debug')

    def build_regex(self):
        return r'geckoview_example\.apk'

    def build_info_regex(self):
        return r'(target|fennec-.*)\.txt'

    def available_bits(self):
        return ()

    def should_use_taskcluster(self):
        # GVE is not on archive.mozilla.org, only on taskcluster
        return True


@REGISTRY.register('fennec-2.3', attr_value='fennec')
class Fennec23Config(FennecConfig):
    tk_name = 'android-api-9'

    def get_nightly_repo_regex(self, date):
        repo = self.get_nightly_repo(date)
        if repo == 'mozilla-central':
            if date < datetime.date(2014, 12, 6):
                repo = "mozilla-central-android"
            else:
                repo = "mozilla-central-android-api-9"
        return self._get_nightly_repo_regex(date, repo)


@REGISTRY.register('jsshell', disable_in_gui=True)
class JsShellConfig(FirefoxConfig):
    def build_info_regex(self):
        # the info file is the one for firefox
        return get_build_regex('firefox', self.os, self.bits, self.processor,
                               with_ext=False) + r'\.txt$'

    def build_regex(self):
        if self.os == 'linux':
            if self.bits == 64:
                part = 'linux-x86_64'
            else:
                part = 'linux-i686'
        elif self.os == 'win':
            if self.bits == 64:
                if self.processor == "aarch64":
                    part = 'win64-aarch64'
                else:
                    part = 'win64(-x86_64)?'
            else:
                part = 'win32'
        else:
            part = 'mac'
        psuffix = '-asan' if 'asan' in self.build_type else ''
        return r'jsshell-%s%s\.zip$' % (part, psuffix)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
from taskcluster import utils as tc_utils

from mozregression.config import (get_defaults, DEFAULT_CONF_FNAME,
                                  TC_CREDENTIALS_FNAME)


def tc_authenticate(logger):
    """
    Returns valid credentials for use with Taskcluster private builds.
    """
    # first, try to load credentials from mozregression config file
    defaults = get_defaults(DEFAULT_CONF_FNAME)
    client_id = defaults.get('taskcluster-clientid')
    access_token = defaults.get('taskcluster-accesstoken')
    if client_id and access_token:
        return dict(clientId=client_id, accessToken=access_token)

    try:
        # else, try to load a valid certificate locally
        with open(TC_CREDENTIALS_FNAME) as f:
            creds = json.load(f)
        if not tc_utils.isExpired(creds['certificate']):
            return creds
    except Exception:
        pass

    # here we need to ask for a certificate, this require web browser
    # authentication
    logger.info(
        "Authentication required from taskcluster. We are going to ask for a"
        " certificate.\nNote that if you have long term access you can instead"
        " set your taskcluster-clientid and taskcluster-accesstoken in the"
        " configuration file (%s)." % DEFAULT_CONF_FNAME
    )
    creds = tc_utils.authenticate("mozregression private build access")

    # save the credentials and the certificate for later use
    with open(TC_CREDENTIALS_FNAME, 'w') as f:
        json.dump(creds, f)
    return creds



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
The BuildInfo classes, that are used to store information a build.
"""
import re
import datetime
from urlparse import urlparse


FIELDS = []


def export(func):
    FIELDS.append(func.__name__)
    return func


class BuildInfo(object):
    """
    Store information about a build to be able to download and run it.

    a BuildInfo is built by calling
    :meth:`mozregression.fetch_build_info.FetchBuildInfo.find_build_info`.
    """
    def __init__(self, fetch_config, build_type, build_url, build_date,
                 changeset, repo_url, repo_name, task_id=None):
        self._fetch_config = fetch_config
        self._build_type = build_type
        self._build_url = build_url
        self._build_date = build_date
        self._changeset = changeset
        self._repo_url = repo_url
        self._repo_name = repo_name
        self._build_file = None
        self._task_id = task_id

    @property
    @export
    def build_type(self):
        """
        Either 'nightly' or 'inbound'
        """
        return self._build_type

    @property
    @export
    def app_name(self):
        """
        The application name, such as "firefox"
        """
        return self._fetch_config.app_name

    @property
    @export
    def build_url(self):
        """
        The url to download the build
        """
        return self._build_url

    @property
    @export
    def build_date(self):
        """
        The date of the build
        """
        return self._build_date

    @property
    @export
    def changeset(self):
        """
        The changeset of the build. For old nightlies, this can be None.
        """
        return self._changeset

    @property
    @export
    def repo_url(self):
        """
        The url of the repository. For old nightlies, this can be None.
        """
        return self._repo_url

    @property
    @export
    def repo_name(self):
        """
        the repository name, e.g. 'mozilla-central'
        """
        return self._repo_name

    @property
    @export
    def build_file(self):
        """
        The build file, once downloaded. This property is None when the
        BuildInfo is instanciated and should be set later.
        """
        return self._build_file

    @property
    @export
    def task_id(self):
        """
        The task ID, for Taskcluster builds.
        """
        return self._task_id

    @build_file.setter
    def build_file(self, build_file):
        self._build_file = build_file

    @property
    def short_changeset(self):
        """
        Returns the first 8 characters of the build changeset.
        """
        return self.changeset[:8]

    def update_from_app_info(self, app_info):
        """
        Takes an app_info (as returned by mozversion) and update the build info
        content if required.

        This helps to build the pushlog url for old nightlies.
        """
        if self._changeset is None:
            self._changeset = app_info.get('application_changeset')
        if self._repo_url is None:
            self._repo_url = app_info.get('application_repository')

    def persist_filename_for(self, data, regex=True):
        """
        Returns the persistent filename for the given data.

        `data` should be a date or datetime object if the build type is
        'nightly', else a changeset.

        if `regex` is True, instead of returning the persistent filename
        it is returned a string (regex pattern) that can match a filename.
        The pattern only allows the build name to be different, by using
        the fetch_config.build_regex() value. For example, it can return:

        '2015-01-11--mozilla-central--firefox.*linux-x86_64\.tar.bz2$'
        """
        if self.build_type == 'nightly':
            if isinstance(data, datetime.datetime):
                prefix = data.strftime("%Y-%m-%d-%H-%M-%S")
            else:
                prefix = str(data)
            persist_part = ''
        else:
            prefix = str(data[:12])
            persist_part = self._fetch_config.inbound_persist_part()
        if persist_part:
            persist_part = '-' + persist_part
        full_prefix = '{}{}--{}--'.format(prefix, persist_part, self.repo_name)
        if regex:
            full_prefix = re.escape(full_prefix)
            appname = self._fetch_config.build_regex()
        else:
            appname = urlparse(self.build_url). \
                path.replace('%2F', '/').split('/')[-1]
        return '{}{}'.format(full_prefix, appname)

    @property
    def persist_filename(self):
        """
        Compute and return the persist filename to use to store this build.
        """
        if self.build_type == 'nightly':
            data = self.build_date
        else:
            data = self.changeset
        return self.persist_filename_for(data, regex=False)

    def to_dict(self):
        """
        Export some public properties as a dict.
        """
        return dict((field, getattr(self, field)) for field in FIELDS)


class NightlyBuildInfo(BuildInfo):
    def __init__(self, fetch_config, build_url, build_date, changeset,
                 repo_url):
        BuildInfo.__init__(self, fetch_config, 'nightly', build_url,
                           build_date, changeset, repo_url,
                           fetch_config.get_nightly_repo(build_date))


class InboundBuildInfo(BuildInfo):
    def __init__(self, fetch_config, build_url, build_date, changeset,
                 repo_url, task_id=None):
        BuildInfo.__init__(self, fetch_config, 'inbound', build_url,
                           build_date, changeset, repo_url,
                           fetch_config.inbound_branch,
                           task_id=task_id)



import datetime

from mozlog import get_proxy_logger

from mozregression.errors import MozRegressionError, EmptyPushlogError
from mozregression.network import retry_get
from mozregression import branches
from mozregression.dates import is_date_or_datetime

LOG = get_proxy_logger("JsonPushes")


class Push(object):
    """
    Simple wrapper around a json push object from json-pushes API.
    """
    __slots__ = ('_data', '_push_id')  # to save memory usage

    def __init__(self, push_id, data):
        self._data = data
        self._push_id = push_id

    @property
    def push_id(self):
        return self._push_id

    @property
    def changesets(self):
        return self._data['changesets']

    @property
    def changeset(self):
        """
        Returns the last changeset in the push (the most interesting for us)
        """
        return self._data['changesets'][-1]

    @property
    def timestamp(self):
        return self._data['date']

    @property
    def utc_date(self):
        return datetime.datetime.utcfromtimestamp(self.timestamp)

    def __str__(self):
        return self.changeset[:12]


class JsonPushes(object):
    """
    Find pushlog Push objects from a mozilla hg json-pushes api.
    """
    def __init__(self, branch='mozilla-inbound'):
        self.branch = branch
        self.repo_url = branches.get_url(branch)

    def pushes(self, **kwargs):
        """
        Returns a sorted lists of Push objects. The list can not be empty.

        Basically issue a raw request to the server.
        """
        base_url = '%s/json-pushes?' % self.repo_url
        url = base_url + '&'.join("%s=%s" % kv for kv in kwargs.iteritems())
        LOG.debug("Using url: %s" % url)

        response = retry_get(url)
        if response.status_code == 404:
            raise MozRegressionError(
                "The url %r returned a 404 error. Please check the"
                " validity of the url." % url
            )
        response.raise_for_status()
        data = response.json()
        if not data:
            raise EmptyPushlogError(
                "The url %r contains no pushlog. Maybe use another range ?"
                % url
            )
        pushlog = []
        for key in sorted(data):
            pushlog.append(Push(key, data[key]))
        return pushlog

    def pushes_within_changes(self, fromchange, tochange, verbose=True,
                              **kwargs):
        """
        Returns a list of Push objects, including fromchange and tochange.

        This will return at least one Push. In case of error it will raise
        a MozRegressionError.
        """
        from_is_date = is_date_or_datetime(fromchange)
        to_is_date = is_date_or_datetime(tochange)

        kwargs = {}
        if not from_is_date:
            # the first changeset is not taken into account in the result.
            # let's add it directly with this request
            chsets = self.pushes(changeset=fromchange)
            kwargs['fromchange'] = fromchange
        else:
            chsets = []
            kwargs['startdate'] = fromchange.strftime('%Y-%m-%d')

        if not to_is_date:
            kwargs['tochange'] = tochange
        else:
            # add one day to take the last day in account
            kwargs['enddate'] = (
                tochange + datetime.timedelta(days=1)).strftime('%Y-%m-%d')

        # now fetch all remaining changesets
        chsets.extend(self.pushes(**kwargs))

        log = LOG.info if verbose else LOG.debug
        if from_is_date:
            first = chsets[0]
            log("Using {} (pushed on {}) for date {}".format(
                first.changeset, first.utc_date, fromchange))
        if to_is_date:
            last = chsets[-1]
            log("Using {} (pushed on {}) for date {}".format(
                last.changeset, last.utc_date, tochange))

        return chsets

    def push(self, changeset, **kwargs):
        """
        Returns the Push object that match the given changeset or date.

        A MozRegressionError is thrown if None is found.
        """
        if is_date_or_datetime(changeset):
            try:
                return self.pushes_within_changes(changeset,
                                                  changeset,
                                                  verbose=False)[-1]
            except EmptyPushlogError:
                raise EmptyPushlogError(
                    "No pushes available for the date %s on %s."
                    % (changeset, self.branch)
                )
        return self.pushes(changeset=changeset, **kwargs)[0]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import stat
import mozfile
from glob import glob
from collections import namedtuple


File = namedtuple('File', ('path', 'stat'))


class PersistLimit(object):
    """
    Keep a list of files, removing the oldest ones when the size_limit
    is reached.

    The access time of a file is used to determine the oldests, e.g. the
    last time a file was read.

    :param size_limit: the size limit in bytes.
    :param file_limit: even if the size limit is reached, this force
                       to keep at least *file_limit* files.
    """
    def __init__(self, size_limit, file_limit=5):
        self.size_limit = size_limit
        self.file_limit = file_limit
        self.files = []
        self._files_size = 0

    def register_file(self, path):
        """
        register a single file.
        """
        try:
            fstat = os.stat(path)
        except OSError:
            # file do not exists probably, just skip it
            # note this happen when backgound files are canceled
            return
        if stat.S_ISREG(fstat.st_mode):
            self.files.append(File(path=path, stat=fstat))
            self._files_size += fstat.st_size

    def register_dir_content(self, directory, pattern="*"):
        """
        Register every files in a directory that match *pattern*.
        """
        for path in glob(os.path.join(directory, pattern)):
            self.register_file(path)

    def remove_old_files(self):
        """
        remove oldest registered files.
        """
        if self.size_limit <= 0 or self.file_limit <= 0:
            return
        # sort by creation time, oldest first
        files = sorted(self.files, key=lambda f: f.stat.st_atime)
        while len(files) > self.file_limit and \
                self._files_size >= self.size_limit:
            f = files.pop(0)
            mozfile.remove(f.path)
            self._files_size -= f.stat.st_size
        self.files = files



import tempfile
import threading
import requests
import os
import sys
import mozfile

from contextlib import closing
from mozlog import get_proxy_logger

from mozregression.persist_limit import PersistLimit

LOG = get_proxy_logger('Download')


class DownloadInterrupt(Exception):
    pass


class Download(object):
    """
    Download is reponsible of downloading one file in the background.

    Example of use: ::

      dl = Download(url, dest)
      dl.start()
      dl.wait() # this will block until completion / cancel / error

    If a download fail or is canceled, the temporary dest is removed from
    the disk.

    :param url: the url of the file to download
    :param dest: the local file path destination
    :param finished_callback: a callback that will be called in the thread
                              when the thread work is done. Takes the download
                              instance as a parameter.
    :param chunk_size: size of the chunk that will be read. The thread can
                        not be stopped while we are reading that chunk size.
    :param session: a requests.Session or the requests module that will do
                    do the real downloading work.
    :param progress: A callable to report the progress (default to None).
                     see :meth:`set_progress`.
    """
    def __init__(self, url, dest, finished_callback=None,
                 chunk_size=16 * 1024, session=requests, progress=None):
        self.thread = threading.Thread(
            target=self._download,
            args=(url, dest, finished_callback, chunk_size, session)
        )
        self._lock = threading.Lock()
        self.__url = url
        self.__dest = dest
        self.__progress = progress
        self.__canceled = False
        self.__error = None

    def start(self):
        """
        Start the thread that will do the download.
        """
        self.thread.start()

    def cancel(self):
        """
        Cancel a previously started download.
        """
        self.__canceled = True

    def is_canceled(self):
        """
        Returns True if we canceled this download.
        """
        return self.__canceled

    def is_running(self):
        """
        Returns True if the downloading thread is running.
        """
        return self.thread.is_alive()

    def wait(self, raise_if_error=True):
        """
        Block until the downloading thread is finished.

        :param raise_if_error: if True (the default), :meth:`raise_if_error`
                               will be called and raise an error if any.
        """
        while self.thread.is_alive():
            try:
                # in case of exception here (like KeyboardInterrupt),
                # cancel the task.
                self.thread.join(0.02)
            except Exception:
                self.cancel()
                raise
        # this will raise exception that may happen inside the thread.
        if raise_if_error:
            self.raise_if_error()

    def error(self):
        """
        Returns None or a tuple of three values (type, value, traceback)
        that give information about the exception.
        """
        return self.__error

    def raise_if_error(self):
        """
        Raise an error if any. If the download was canceled, raise
        :class:`DownloadInterrupt`.
        """
        if self.__error:
            raise self.__error[0], self.__error[1], self.__error[2]
        if self.__canceled:
            raise DownloadInterrupt()

    def set_progress(self, progress):
        """
        set a callable to report the progress of the download, or None to
        disable any report.

        The callable must take three parameters (download, current, total).
        Note that this method is thread safe, you can call it during a
        download.
        """
        with self._lock:
            self.__progress = progress

    def get_dest(self):
        """
        Returns the dest.
        """
        return self.__dest

    def get_url(self):
        """
        Returns the url.
        """
        return self.__url

    def _update_progress(self, current, total):
        with self._lock:
            if self.__progress:
                self.__progress(self, current, total)

    def _download(self, url, dest, finished_callback, chunk_size, session):
        # save the file under a temporary name
        # this allow to not use a broken file in case things went really bad
        # while downloading the file (ie the python interpreter is killed
        # abruptly)
        temp = None
        bytes_so_far = 0
        try:
            with closing(session.get(url, stream=True)) as response:
                total_size = int(response.headers['Content-length'].strip())
                self._update_progress(bytes_so_far, total_size)
                # we use NamedTemporaryFile as raw open() call was causing
                # issues on windows - see:
                # https://bugzilla.mozilla.org/show_bug.cgi?id=1185756
                with tempfile.NamedTemporaryFile(
                        delete=False,
                        suffix='.tmp',
                        dir=os.path.dirname(dest)) as temp:
                    for chunk in response.iter_content(chunk_size):
                        if self.is_canceled():
                            break
                        if chunk:
                            temp.write(chunk)
                        bytes_so_far += len(chunk)
                        self._update_progress(bytes_so_far, total_size)
            response.raise_for_status()
        except Exception:
            self.__error = sys.exc_info()
        try:
            if temp is None:
                pass  # not even opened the temp file, nothing to do
            elif self.is_canceled() or self.__error:
                mozfile.remove(temp.name)
            else:
                # if all goes well, then rename the file to the real dest
                mozfile.remove(dest)  # just in case it already existed
                mozfile.move(temp.name, dest)
        finally:
            if finished_callback:
                finished_callback(self)


class DownloadManager(object):
    """
    DownloadManager is responsible of starting and managing downloads inside
    a given directory. It will download a file only if a given filename
    is not already there.

    Downloadmanager itself is not thread safe, and must not be shared
    between threads.

    Note that backgound downloads needs to be stopped. For example, if
    you have an exception while a download is occuring, python will only
    exit when the download will finish. To get rid of that, there is a
    possible idiom: ::

      def download_things(manager):
          # do things with the manager
          manager.download(url1, f1)
          manager.download(url2, f2)
          ...

      manager = DownloadManager(destdir)
      try:
          download_things(manager)
      finally:
          # ensure we cancel all background downloads to ask the end
          # of possible remainings threads
          manager.cancel()

    :param destdir: a directory where files are downloaded. It is required
                    that it exists.
    :param session: a requests session
    :param persist_limit: an instance of :class:`PersistLimit`, to allow
                          limiting the size of the download dir. Defaults
                          to None, meaning no limit.
    """
    def __init__(self, destdir, session=requests, persist_limit=None):
        self.destdir = destdir
        self.session = session
        self._downloads = {}
        self._lock = threading.Lock()
        self.persist_limit = persist_limit or PersistLimit(0)
        self.persist_limit.register_dir_content(self.destdir)

        # if persist folder does not exist, create it
        if not os.path.isdir(destdir):
            os.makedirs(destdir)

    def get_dest(self, fname):
        return os.path.join(self.destdir, fname)

    def cancel(self, cancel_if=None):
        """
        Cancel downloads, if any.

        if cancel_if is given, it must be a callable that take the download
        instance as parameter, and return True if the download needs to be
        canceled.

        Note that download threads won't be stopped directly.
        """
        with self._lock:
            for download in self._downloads.itervalues():
                if cancel_if is None or cancel_if(download):
                    if download.is_running():
                        download.cancel()

    def wait(self, raise_if_error=True):
        """
        Wait for all downloads to be finished.
        """
        for download in self._downloads.values():
            download.wait(raise_if_error=raise_if_error)

    def download(self, url, fname):
        """
        Returns a started download instance, or None if fname is already
        present in destdir.

        if a download is already running for the given fname, it is just
        returned. Else the download is created, started and returned.
        """
        dest = self.get_dest(fname)
        with self._lock:
            # if we are downloading, just returns the instance
            if dest in self._downloads:
                return self._downloads[dest]

        if os.path.exists(dest):
            return None

        # else create the download (will be automatically removed of
        # the list on completion) start it, and returns that.
        with self._lock:
            download = Download(url, dest,
                                session=self.session,
                                finished_callback=self._download_finished)
            self._downloads[dest] = download
            download.start()
            self._download_started(download)
            return download

    def _download_started(self, _):
        pass

    def _download_finished(self, dl):
        with self._lock:
            dest = dl.get_dest()
            del self._downloads[dest]
            self.persist_limit.register_file(dest)
            self.persist_limit.remove_old_files()


def download_progress(_dl, bytes_so_far, total_size):
    percent = (float(bytes_so_far) / total_size) * 100
    sys.stdout.write("===== Downloaded %d%% =====\r" % percent)
    sys.stdout.flush()


class BuildDownloadManager(DownloadManager):
    """
    A DownloadManager specialized to download builds.
    """
    def __init__(self, destdir, session=requests,
                 background_dl_policy='cancel',
                 persist_limit=None):
        DownloadManager.__init__(self, destdir, session=session,
                                 persist_limit=persist_limit)
        self._downloads_bg = set()
        assert background_dl_policy in ('cancel', 'keep')
        self.background_dl_policy = background_dl_policy

    def _extract_download_info(self, build_info):
        return build_info.build_url, build_info.persist_filename

    def download_in_background(self, build_info):
        """
        Start a build download in background.

        Don nothing is a build is already downloading/downloaded.
        """
        build_url, fname = self._extract_download_info(build_info)
        result = self.download(build_url, fname)
        if result is not None:
            self._downloads_bg.add(fname)
        return result

    def focus_download(self, build_info):
        """
        Start a download for a build and focus on it.

        *focus* here means that if there are running downloads for other
        builds they will be canceled. Also, the progress is attached so
        the user can see the download progress.

        If the download of the build is already running, it will just
        attach the progress function. If the build has already been
        downloaded, it will do nothing.

        this methods block until the build is available, or any error
        occurs.

        Returns the complete path of the downloaded build. Also this path
        is set for the `build_info.build_file` property.
        """
        build_url, fname = self._extract_download_info(build_info)
        dest = self.get_dest(fname)
        build_info.build_file = dest
        # first, stop all downloads in background (except the one for this
        # build if any)
        if self.background_dl_policy == 'cancel':
            self.cancel(cancel_if=lambda dl: dest != dl.get_dest())

        dl = self.download(build_url, fname)
        if dl:
            LOG.info("Downloading build from: %s" % build_url)
            dl.set_progress(download_progress)
            try:
                dl.wait()
            finally:
                print ''  # a new line after download_progress calls

        else:
            msg = "Using local file: %s" % dest
            if fname in self._downloads_bg:
                msg += " (downloaded in background)"
            LOG.info(msg)
        return dest



import sys
from setuptools import setup
from mozregression import __version__
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """
    Run py.test with the "python setup.py test command"
    """
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.pytest_args += (' ' + self.distribution.test_suite)

    def run_tests(self):
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


if sys.version_info < (2, 7) or sys.version_info >= (3, 0):
    sys.exit("mozregression currently require python >=2.7 and <3.")

# we pin these dependencies in the requirements files -- all of these
# should be python 3 compatible
DEPENDENCIES = [
    'beautifulsoup4>=4.7.1',
    'colorama>=0.4.1',
    'configobj>=5.0.6',
    'mozdevice>=3.0.1',
    'mozfile>=2.0.0',
    'mozinfo>=1.1.0',
    'mozinstall>=2.0.0',
    'mozlog>=4.0',
    'mozprocess>=1.0.0',
    'mozprofile>=2.2.0',
    'mozrunner>=7.3.0',
    'mozversion>=2.1.0',
    'redo>=2.0.2',
    'requests[security]>=2.21.0',
    'taskcluster>=6.0.0',
]

desc = """Regression range finder for Mozilla nightly builds"""
long_desc = """Regression range finder for Mozilla nightly builds.
For more information see the mozregression website:
http://mozilla.github.io/mozregression/"""

setup(name="mozregression",
      version=__version__,
      description=desc,
      long_description=long_desc,
      author='Mozilla Automation and Tools Team',
      author_email='tools@lists.mozilla.org',
      url='http://github.com/mozilla/mozregression',
      license='MPL 1.1/GPL 2.0/LGPL 2.1',
      packages=['mozregression'],
      entry_points="""
          [console_scripts]
          mozregression = mozregression.main:main
        """,
      platforms=['Any'],
      install_requires=DEPENDENCIES,
      tests_require=['mock', 'pytest', 'pytest-mock'],
      test_suite='tests',
      cmdclass={'test': PyTest},
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Operating System :: OS Independent'
      ])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
This module provides an API to integrate mozregression within mach.

Be careful to not break backward compatibility here, or you will
break mach!
"""

from argparse import Namespace

from mozregression.cli import create_parser
from mozregression.config import DEFAULT_CONF_FNAME, get_defaults
from mozregression.main import main, pypi_latest_version
from mozregression import __version__


def new_release_on_pypi():
    """
    Check if a new release is available on pypi and returns it.

    None is returned in case of error or if there is no new version.
    """
    try:
        pypi_version = pypi_latest_version()
    except Exception:
        return
    if pypi_version != __version__:
        return pypi_version


def parser():
    """
    Create and returns the mozregression ArgumentParser instance.
    """
    defaults = get_defaults(DEFAULT_CONF_FNAME)
    return create_parser(defaults=defaults)


def run(options):
    """
    Run mozregression given a dict of options.
    """
    main(namespace=Namespace(**options), check_new_version=False)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
This module implements a :class:`TestRunner` interface for testing builds
and a default implementation :class:`ManualTestRunner`.
"""

from mozlog import get_proxy_logger
import subprocess
import shlex
import os
import datetime

from mozregression.launchers import create_launcher as mozlauncher
from mozregression.errors import TestCommandError, LauncherError
from abc import ABCMeta, abstractmethod

LOG = get_proxy_logger("Test Runner")


def create_launcher(build_info):
    """
    Create and returns a :class:`mozregression.launchers.Launcher`.
    """
    if build_info.build_type == 'nightly':
        if isinstance(build_info.build_date, datetime.datetime):
            desc = ("for buildid %s"
                    % build_info.build_date.strftime("%Y%m%d%H%M%S"))
        else:
            desc = "for %s" % build_info.build_date
    else:
        desc = ("built on %s, revision %s"
                % (build_info.build_date,
                   build_info.short_changeset))
    LOG.info("Running %s build %s" % (build_info.repo_name, desc))

    return mozlauncher(build_info)


class TestRunner:
    """
    Abstract class that allows to test a build.

    :meth:`evaluate` must be implemented by subclasses.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def evaluate(self, build_info, allow_back=False):
        """
        Evaluate a given build. Must returns a tuple of (verdict, app_info).

        The verdict must be a letter that indicate the state of the build:
        'g', 'b', 's', 'r' or 'e' respectively for 'good', 'bad', 'skip',
        'retry' or 'exit'. If **allow_back** is True, it is also possible
        to return 'back'.

        The app_info is the return value of the
        :meth:`mozregression.launchers.Launcher.get_app_info` for this
        particular build.

        :param build_path: the path to the build file to test
        :param build_info: a :class:`mozrgression.uild_info.BuildInfo` instance
        :param allow_back: indicate if the back command should be proposed.
        """
        raise NotImplementedError

    @abstractmethod
    def run_once(self, build_info):
        """
        Run the given build and wait for its completion. Return the error
        code when available.
        """
        raise NotImplementedError

    def index_to_try_after_skip(self, build_range):
        """
        Return the index of the build to use after a build was skipped.

        By default this only returns the mid point of the remaining range.
        """
        return build_range.mid_point()


class ManualTestRunner(TestRunner):
    """
    A TestRunner subclass that run builds and ask for evaluation by
    prompting in the terminal.
    """
    def __init__(self, launcher_kwargs=None):
        TestRunner.__init__(self)
        self.launcher_kwargs = launcher_kwargs or {}

    def get_verdict(self, build_info, allow_back):
        """
        Ask and returns the verdict.
        """
        options = ['good', 'bad', 'skip', 'retry', 'exit']
        if allow_back:
            options.insert(-1, 'back')
        # allow user to just type one letter
        allowed_inputs = options + [o[0] for o in options]
        # format options to nice printing
        formatted_options = (', '.join(["'%s'" % o for o in options[:-1]]) +
                             " or '%s'" % options[-1])
        verdict = ""
        while verdict not in allowed_inputs:
            verdict = raw_input("Was this %s build good, bad, or broken?"
                                " (type %s and press Enter): "
                                % (build_info.build_type,
                                   formatted_options))

        if verdict == 'back':
            return 'back'
        # shorten verdict to one character for processing...
        return verdict[0]

    def evaluate(self, build_info, allow_back=False):
        with create_launcher(build_info) as launcher:
            launcher.start(**self.launcher_kwargs)
            build_info.update_from_app_info(launcher.get_app_info())
            verdict = self.get_verdict(build_info, allow_back)
            try:
                launcher.stop()
            except LauncherError:
                # we got an error on process termination, but user
                # already gave the verdict, so pass this "silently"
                # (it would be logged from the launcher anyway)
                launcher._running = False
        return verdict

    def run_once(self, build_info):
        with create_launcher(build_info) as launcher:
            launcher.start(**self.launcher_kwargs)
            build_info.update_from_app_info(launcher.get_app_info())
            return launcher.wait()

    def index_to_try_after_skip(self, build_range):
        mid = TestRunner.index_to_try_after_skip(self, build_range)
        build_range_len = len(build_range)
        if build_range_len <= 3:
            # do not even ask if there is only one build to choose
            return mid
        min = -mid + 1
        max = build_range_len - mid - 2
        valid_range = range(min, max + 1)
        print("Build was skipped. You can manually choose a new build to"
              " test, to be able to get out of a broken build range.")
        print("Please type the index of the build you would like to try - the"
              " index is 0-based on the middle of the remaining build range.")
        print "You can choose a build index between [%d, %d]:" % (min, max)
        while True:
            value = raw_input('> ')
            try:
                index = int(value)
                if index in valid_range:
                    return mid + index
            except ValueError:
                pass


def _raise_command_error(exc, msg=''):
    raise TestCommandError("Unable to run the test command%s: `%s`"
                           % (msg, exc))


class CommandTestRunner(TestRunner):
    """
    A TestRunner subclass that evaluate builds given a shell command.

    Some variables may be used to evaluate the builds:
     - variables referenced in :meth:`TestRunner.evaluate`
     - app_name (the tested application name: firefox, ...)
     - binary (the path to the binary when applicable - not for fennec)

    These variables can be used in two ways:
    1. as environment variables. 'MOZREGRESSION_' is prepended and the
       variables names are upcased. Example: MOZREGRESSION_BINARY
    2. as placeholders in the command line. variables names must be enclosed
       with curly brackets. Example:
       `mozmill -app firefox -b {binary} -t path/to/test.js`
    """
    def __init__(self, command):
        TestRunner.__init__(self)
        self.command = command

    def evaluate(self, build_info, allow_back=False):
        with create_launcher(build_info) as launcher:
            build_info.update_from_app_info(launcher.get_app_info())
            variables = {k: v for k, v in build_info.to_dict().iteritems()}
            if hasattr(launcher, 'binary'):
                variables['binary'] = launcher.binary

            env = dict(os.environ)
            for k, v in variables.iteritems():
                if type(v) == unicode:
                    v = v.encode('ascii', errors='ignore')
                else:
                    env['MOZREGRESSION_' + k.upper()] = str(v)
            try:
                command = self.command.format(**variables)
            except KeyError as exc:
                _raise_command_error(exc, ' (formatting error)')
            LOG.info('Running test command: `%s`' % command)
            cmdlist = shlex.split(command)
            try:
                retcode = subprocess.call(cmdlist, env=env)
            except IndexError:
                _raise_command_error("Empty command")
            except OSError as exc:
                _raise_command_error(exc,
                                     " (%s not found or not executable)"
                                     % cmdlist[0])
        LOG.info('Test command result: %d (build is %s)'
                 % (retcode, 'good' if retcode == 0 else 'bad'))
        return 'g' if retcode == 0 else 'b'

    def run_once(self, build_info):
        return 0 if self.evaluate(build_info) == 'g' else 1
